/* -*-C-*-
*******************************************************************************
*
* File:         chipper.c
* RCS:          $Header: $
* Description:  CHIPPER, translated from the original Turbo Pascal source code
* Author:       Darryl Okahata (darrylo@sr.hp.com)
* Created:      Sun Jul  7 02:17:25 1991
* Modified:     Mon Jul  8 19:46:46 1991 (Darryl Okahata) darrylo@hpsrdmo
* Language:     C
* Package:      CHIPPER
*
*******************************************************************************
*
* This C source file is a "~99%-direct" translation of Christian Egeberg's
* Chip-48 assembler from the original Turbo Pascal source.  The bugs in the
* original version are still in this one, as I never tried to fix them (in
* particular, if you specify files on the command line, make sure that you
* specify *all* of the files, not just one or two).  The vast majority of the
* translation was done using Dave Gillespie's excellent Pascal-to-C translator
* (version 1.20), needing only a few tweaks, bugfixes, and (Turbo Pascal)
* support routines.  What few comments were in the original Pascal code were
* preserved by the translator, so please don't complain to me about the lack
* of comments (note that the translator added some warning comments, which
* I've left in).  For the most part, the translated code is readable (it looks
* VERY similar to the original Pascal code), but not very efficient.  String
* handling is particularly inefficient, although the speed is quite acceptable
* on my workstation.
*
* This code was compiled and lightly tested under Unix, although trivial and
* TOTALLY UNTESTED #ifdefs were added for MSDOS (just #define MSDOS while
* compiling).  The translator was told to assume 16-bit ints on the target
* machine when doing the translating, so this should make an MSDOS port even
* easier (fortunately, this does not seem to cause any problems with a 32-bit
* Unix version).  Note that this program will require more than the default
* 2-4K of stack space allocated by MSDOS compilers; taking a wild guess, 16K
* might be enough (use the large memory model!).
*
* Note that this code has not been run through lint(1).  It should be, but I'm
* lazy.
*
* For those people who want to use p2c to re-translate the code, here are a
* few tips and notes:
*
*    1. Most of the problems with the translated code are obvious -- the code
*	will not compile because of them.
*    2. You must tell the translator to not use small sets, by adding the
*	following line to the local p2c configuration file:
*
*		SmallSetConst		-1
*
*	Using sets of characters seems to cause heartburn for p2c version
*	1.20, and it's best to not have it use small sets, as the non-small-
*	set code is easier to correct (yes, it's still incorrect).
*    3. The code to initialize the "Operators" set is incorrect (the compiler
*	initialization of this variable uses a non-constant expression).  To
*	have it properly initialized, move the initialization code to the
*	beginning of main() (actually, the code is correct -- it's just in the
*	"wrong place").
*    4. The above-mentioned initialization code refers to a variable called
*	"SET", declared to be "long SET[2]".  This is incorrect; change it to
*	"long SET[5]".
*
*******************************************************************************
*/


#ifndef MSDOS
/*
 * Unix definitions:
 */
# include <limits.h>
# ifndef PATH_MAX
#  define PATH_MAX	1024		/* pray */
# endif
# define CHIP_PATH_LEN	(PATH_MAX + 1)
# define CHIP_NAME_LEN	(PATH_MAX + 1)	/* A bit wasteful */
# define CHIP_EXT_LEN	(PATH_MAX + 1)	/* Utterly wasteful, but it's easy */
#else
/*
 * MSDOS definitions:
 */
# define CHIP_PATH_LEN	80		/* Is this really correct??? */
# define CHIP_NAME_LEN	9
# define CHIP_EXT_LEN	4
#endif


/*****************************************************************************/
/* Output from p2c, the Pascal-to-C translator */
/* From input file "chipper.pas" */


/* Chip-48 Assembler V1.12 by Christian Egeberg 2/11-'90 .. 7/11-'90 */


#include "p2c.h"


/*$R-,S+,I+,F-,O-,A+,V+,B-,N-,E-,D-,L-,M 16384,16384,655360*/


#define CopyRight       "Chip-48 Assembler V1.12 by Christian Egeberg 7/11-'90"

#define ErrorExitCode   1
#define StartAddress    0x200
#define StopAddress     0xfff

#define WordMask        0xffffL

#define AddrMask        0xfff
#define ByteMask        0xff
#define NibbleMask      0xf
#define LineLength      160
#define ParamLength     80
#define SymbolLength    16
#define MaxParams       64

#define NullChar        '\0'
#define BellChar        '\007'
#define SpaceChar       ' '
#define SeparatorChar   ','
#define RemarkChar      ';'
#define SymbolChar      '_'
#define LabelChar       ':'
#define EqualChar       '='
#define TextChar        '\''
#define AddressChar     '?'
#define HexChar         '#'
#define BinChar         '$'
#define OctChar         '@'
#define AscChar         '"'
#define StartChar       '('
#define StopChar        ')'
#define PlusChar        '+'
#define MinusChar       '-'
#define NotChar         '~'
#define PowerChar       '!'
#define ShlChar         '<'
#define ShrChar         '>'
#define MulChar         '*'
#define FracChar        '/'
#define AndChar         '&'
#define OrChar          '|'
#define XorChar         '^'
#define DivChar         '\\'
#define ModChar         '%'

#define NameDefault     ""
#ifndef MSDOS
# define InExtDefault    ".chp"
# define ListExtDefault  ".lst"
# define OutExtDefault   ".bin"
#else
# define InExtDefault    ".CHP"
# define ListExtDefault  ".LST"
# define OutExtDefault   "."
#endif

#define RunErrorMessage  "Fatal error: "
#define RunWarningMessage  "Warning: "
#define WarningNumMessage  "Total number of warnings: "
#define NoSourceError   "No source file found"
#define FileAccessError  "Unable to open file"
#define BoundsError     "Outside legal address range"
#define ParamCountWarning  "Incorrect number of parameters"
#define DualSymbolWarning  "No directive found"
#define NoSymbolWarning  "No symbol associated"
#define CopySymbolWarning  "Attempt to redefine existing symbol"
#define UndefinedWarning  "Badly defined parameter"
#define RangeWarning    "Parameter out of range"
#define NoRegisterWarning  "Register not found"
#define BadRegisterWarning  "Illegal register"
#define InternalWarning  "Internal data structure mismatch"


typedef enum {
    EqualToken, AddToken, AndToken, CallToken, ClsToken, DaToken, DbToken,
    DrwToken, DsToken, DwToken, EndToken, EquToken, IncludeToken, JpToken,
    LdToken, OrToken, OrgToken, RetToken, RndToken, SeToken, ShlToken,
    ShrToken, SknpToken, SkpToken, SneToken, SubToken, SubnToken, SysToken,
    XorToken, LastToken
} Token;
typedef enum {
    BReg, DtReg, FReg, IReg, KReg, V0Reg, V1Reg, V2Reg, V3Reg, V4Reg, V5Reg,
    V6Reg, V7Reg, V8Reg, V9Reg, VaReg, VbReg, VcReg, VdReg, VeReg, VfReg,
    StReg, IiReg, LastReg
} Register_;
typedef long CharSet[9];

typedef Char LineString[LineLength + 1];

typedef Char ParamString[ParamLength + 1];

typedef Char SymbolString[SymbolLength + 1];

typedef struct ParamRecord {
    ParamString Param;
    struct ParamRecord *Next;
} ParamRecord;

typedef struct SymbolRecord {
    SymbolString Symbol;
    long Address;
    struct SymbolRecord *Left, *Right;
} SymbolRecord;

typedef struct InstRecord {
    unsigned short Line;
    Char Name[CHIP_PATH_LEN];
    long Address;
    Token Inst;
    uchar Count;
    ParamRecord *Params;
    struct InstRecord *Next, *Prev;
} InstRecord;
/* p2c: chipper.pas, line 119: Warning: Expected an ordinal type [156] */


Static SymbolString TokenText[30] = {
    "=", "ADD", "AND", "CALL", "CLS", "DA", "DB",
    "DRW", "DS", "DW", "END", "EQU", "INCLUDE", "JP", "LD", "OR", "ORG",
    "RET", "RND", "SE", "SHL", "SHR", "SKNP", "SKP", "SNE", "SUB", "SUBN",
    "SYS", "XOR", ""
};

Char STR1[256];

Static SymbolString RegisterText[24] = {
    "B", "DT", "F", "I", "K", "V0", "V1", "V2", "V3", "V4", "V5", "V6", "V7",
    "V8", "V9", "VA", "VB", "VC", "VD", "VE", "VF", "ST", "[I]", ""
};

Static long	*Operators;
/* p2c: chipper.pas, line 144: Warning: Expected an ordinal type [156] */
long SET[5];
Static ParamString DigitText = "0123456789ABCDEF";
Static InstRecord *Instructions = NULL;
Static SymbolRecord *Directives = NULL, *Registers_ = NULL, *Symbols = NULL,
		    *LastSymbol = NULL;
Static long Current = StartAddress, Finish = StartAddress;
Static LineString LineText = "";
Static Char LineName[CHIP_PATH_LEN] = NameDefault;
Static unsigned short LineNum = 1;
Static InstRecord *InstPoint = NULL;
Static boolean ListOpen = false;
Static unsigned short WarningCount = 0;

Static uchar Memory[StopAddress - StartAddress + 1];
Static FILE *OutFile;
Static FILE *StdIn, *StdOut, *ListFile;
Static Char InFileName[CHIP_PATH_LEN], OutFileName[CHIP_PATH_LEN],
	    ListFileName[CHIP_PATH_LEN], FileDir[CHIP_PATH_LEN];
Static Char FileName[CHIP_NAME_LEN];
Static Char FileExt[CHIP_EXT_LEN];
Static Char OutFile_NAME[_FNSIZE];
Static Char StdIn_NAME[_FNSIZE];
Static Char StdOut_NAME[_FNSIZE];
Static Char ListFile_NAME[_FNSIZE];


Static Void RunError(Message)
Char *Message;
{
    SymbolString Number;
    ParamRecord *Param;

    fprintf(StdOut, "%c%s%s\n", BellChar, RunErrorMessage, Message);
    if (ListOpen)
	fprintf(ListFile, "%s%s\n", RunErrorMessage, Message);
    if (strcmp(LineName, NameDefault)) {
	sprintf(Number, "%u", LineNum);
/* p2c: chipper.pas, line 156:
 * Note: Possible string truncation in assignment [145] */
	fprintf(StdOut, "Current file %s line %s\n", LineName, Number);
	fprintf(StdOut, "%s\n", LineText);
	if (ListOpen) {
	    fprintf(ListFile, "Current file %s line %s\n", LineName, Number);
	    fprintf(ListFile, "%s\n", LineText);
	}
    }
    if (InstPoint != NULL) {
	sprintf(Number, "%u", InstPoint->Line);
/* p2c: chipper.pas, line 167:
 * Note: Possible string truncation in assignment [145] */
	fprintf(StdOut, "Associated file %s line %s\n",
		InstPoint->Name, Number);
	fputs(TokenText[(long)InstPoint->Inst], StdOut);
	if (ListOpen) {
	    fprintf(ListFile, "Associated file %s line %s\n",
		    InstPoint->Name, Number);
	    fputs(TokenText[(long)InstPoint->Inst], ListFile);
	}
	Param = InstPoint->Params;
	while (Param != NULL) {
	    fprintf(StdOut, ", %s", Param->Param);
	    if (ListOpen)
		fprintf(ListFile, ", %s", Param->Param);
	    Param = Param->Next;
	}
	putc('\n', StdOut);
	if (ListOpen)
	    putc('\n', ListFile);
    }
    putc('\n', StdOut);
    if (ListOpen)
	putc('\n', ListFile);
    _Escape(ErrorExitCode);
    /* Turbo Pascal closes all files on exit */
}


Static Void RunWarning(Message)
Char *Message;
{
    SymbolString Number;
    ParamRecord *Param;

    fprintf(StdOut, "%s%s\n", RunWarningMessage, Message);
    fprintf(ListFile, "%s%s\n", RunWarningMessage, Message);
    if (strcmp(LineName, NameDefault)) {
	sprintf(Number, "%u", LineNum);
/* p2c: chipper.pas, line 207:
 * Note: Possible string truncation in assignment [145] */
	fprintf(StdOut, "Current file %s line %s\n", LineName, Number);
	fprintf(StdOut, "%s\n", LineText);
	fprintf(ListFile, "Current file %s line %s\n", LineName, Number);
	fprintf(ListFile, "%s\n", LineText);
    }
    if (InstPoint != NULL) {
	sprintf(Number, "%u", InstPoint->Line);
/* p2c: chipper.pas, line 215:
 * Note: Possible string truncation in assignment [145] */
	fprintf(StdOut, "Associated file %s line %s\n",
		InstPoint->Name, Number);
	fputs(TokenText[(long)InstPoint->Inst], StdOut);
	fprintf(ListFile, "Associated file %s line %s\n",
		InstPoint->Name, Number);
	fputs(TokenText[(long)InstPoint->Inst], ListFile);
	Param = InstPoint->Params;
	while (Param != NULL) {
	    fprintf(StdOut, ", %s", Param->Param);
	    fprintf(ListFile, ", %s", Param->Param);
	    Param = Param->Next;
	}
	putc('\n', StdOut);
	putc('\n', ListFile);
    }
    putc('\n', StdOut);
    putc('\n', ListFile);
    WarningCount++;
}


Static Char *HexString(Result, Value, Count)
Char *Result;
long Value;
uchar Count;
{
    SymbolString FoundWord;
    uchar Digit;
    Char STR2[18];

    *FoundWord = '\0';
    while (Value > 0) {
	Digit = Value & NibbleMask;
	sprintf(FoundWord, "%c%s", DigitText[Digit], strcpy(STR2, FoundWord));
/* p2c: chipper.pas, line 248:
 * Note: Possible string truncation in assignment [145] */
	Value /= 16;
    }
    while (strlen(FoundWord) < Count) {
	sprintf(FoundWord, "0%s", strcpy(STR2, FoundWord));
/* p2c: chipper.pas, line 252:
 * Note: Possible string truncation in assignment [145] */
    }
    return (strcpy(Result, FoundWord));
}


Static Void ListInstruction(Address, Count, Inst, Number)
unsigned short Address, Count;
InstRecord *Inst;
uchar Number;
{
    ParamRecord *Param;
    unsigned short This;
    SymbolString STR3;

    fprintf(ListFile, "%5u %s", Inst->Line, TokenText[(long)Inst->Inst]);
    Param = Inst->Params;
    for (This = 1; This <= Number; This++) {
	if (Param != NULL) {
	    fprintf(ListFile, ", %s", Param->Param);
	    Param = Param->Next;
	}
    }
    fprintf(ListFile, "\n%s ", HexString(STR3, (long)Address, 3));
    for (This = 0; This < Count; This++)
	fputs(HexString(STR3, (long)Memory[Address + This - StartAddress], 2),
	      ListFile);
    fprintf(ListFile, "\n\n");
}


Static Void ListSymbols(Head)
SymbolRecord *Head;
{
    SymbolString STR2;

    if (Head->Left != NULL)
	ListSymbols(Head->Left);
    fprintf(ListFile, "%s %s\n",
	    HexString(STR2, Head->Address, 3), Head->Symbol);
    if (Head->Right != NULL)
	ListSymbols(Head->Right);
}


Static Void ListWarnings()
{
    fprintf(StdOut, "%s%u\n", WarningNumMessage, WarningCount);
    fprintf(ListFile, "\n%s%u\n", WarningNumMessage, WarningCount);
}


Static Char *ExpandFileName(Result, Path, DefName, DefExt)
Char *Result, *Path;
Char *DefName;
Char *DefExt;
{
    Char Dir[CHIP_PATH_LEN];
    Char Name[CHIP_NAME_LEN];
    Char Ext[CHIP_EXT_LEN];
    Char STR2[256];

/* p2c: chipper.pas, line 307:
 * Warning: Symbol 'FEXPAND' is not defined [221] */
/* p2c: chipper.pas, line 307:
 * Warning: Mixing non-strings with strings [170] */
    FSplit(FExpand(Path), Dir, Name, Ext);
    if (*Name == '\0')
	strcpy(Name, DefName);
    if (*Ext == '\0')
	strcpy(Ext, DefExt);
    strcpy(FileDir, Dir);
    strcpy(FileName, Name);
    strcpy(FileExt, Ext);
    sprintf(Result, "%s%s%s", Dir, Name, Ext);
    return (Result);
/* p2c: chipper.pas, line 315:
 * Note: Possible string truncation in assignment [145] */
}


Static boolean OpenFiles()
{
    boolean Result;
    Char STR2[CHIP_PATH_LEN];
    Char STR3[256];
    Char *TEMP;

    Result = false;
    if (P_argc >= 2) {
	ExpandFileName(InFileName, strcpy(STR3, P_argv[1]), NameDefault,
		       InExtDefault);
	if (strcmp(FileName, NameDefault)) {
	    strcpy(OutFileName, P_argv[2]);
/* p2c: chipper.pas, line 330:
 * Note: Possible string truncation in assignment [145] */
	    strcpy(ListFileName, P_argv[3]);
/* p2c: chipper.pas, line 331:
 * Note: Possible string truncation in assignment [145] */
	} else
	    RunError(NoSourceError);
    } else {
	fprintf(StdOut, "SourceFileName? ");
	fgets(InFileName, CHIP_PATH_LEN, StdIn);
	TEMP = strchr(InFileName, '\n');
	if (TEMP != NULL)
	    *TEMP = 0;
	strcpy(InFileName,
	       ExpandFileName(STR2, InFileName, NameDefault, InExtDefault));
	if (strcmp(FileName, NameDefault)) {
	    fprintf(StdOut, "TargetFileName? ");
	    fgets(OutFileName, CHIP_PATH_LEN, StdIn);
	    TEMP = strchr(OutFileName, '\n');
	    if (TEMP != NULL)
		*TEMP = 0;
	    fprintf(StdOut, "ListFileName? ");
	    fgets(ListFileName, CHIP_PATH_LEN, StdIn);
	    TEMP = strchr(ListFileName, '\n');
	    if (TEMP != NULL)
		*TEMP = 0;
	    putc('\n', StdOut);
	} else
	    RunError(NoSourceError);
    }
    sprintf(STR3, "%s", OutExtDefault);
    strcpy(OutFileName, ExpandFileName(STR2, OutFileName, FileName, STR3));
    strcpy(ListFileName,
	   ExpandFileName(STR2, ListFileName, FileName, ListExtDefault));
    fprintf(StdOut, "SourceFile: %s\n", InFileName);
    fprintf(StdOut, "TargetFile: %s\n", OutFileName);
    strcpy(OutFile_NAME, OutFileName);
    if (OutFile != NULL)
	OutFile = freopen(OutFile_NAME, "w+b", OutFile);
    else
	OutFile = fopen(OutFile_NAME, "w+b");
    _SETIO(OutFile != NULL, FileNotFound);
    if (P_ioresult != 0)
	RunError(FileAccessError);
    fprintf(StdOut, "ListFile: %s\n", ListFileName);
    strcpy(ListFile_NAME, ListFileName);
    if (ListFile != NULL)
	ListFile = freopen(ListFile_NAME, "w", ListFile);
    else
	ListFile = fopen(ListFile_NAME, "w");
    _SETIO(ListFile != NULL, FileNotFound);
    if (P_ioresult != 0)
	RunError(FileAccessError);
    putc('\n', StdOut);
    ListOpen = true;
    fprintf(ListFile, "%s\n\n", CopyRight);
    fprintf(ListFile, "SourceFile: %s\n", InFileName);
    fprintf(ListFile, "TargetFile: %s\n", OutFileName);
    fprintf(ListFile, "ListFile: %s\n\n", ListFileName);
    return (true);
}


Static Char *StripSymbol(Result, Symbol_)
Char *Result, *Symbol_;
{
    SymbolString Symbol;
    Char STR2[256];

    strcpy(Symbol, Symbol_);
    if (Symbol[0] == SymbolChar) {
	strcpy(Symbol, strsub(STR2, Symbol, 2, (int)(strlen(Symbol) - 1)));
/* p2c: chipper.pas, line 384:
 * Note: Possible string truncation in assignment [145] */
    }
    if (Symbol[strlen(Symbol) - 1] == LabelChar)
	Symbol[strlen(Symbol) - 1] = '\0';
    return (strcpy(Result, Symbol));
}


Static boolean DefineSymbol(Symbol_, Value, Head)
Char *Symbol_;
long Value;
SymbolRecord **Head;
{
    boolean Result;
    SymbolString Symbol;
    SymbolRecord *Symb, *Last;
    SymbolString STR2;

    strcpy(Symbol, Symbol_);
    Result = true;
    strcpy(Symbol, StripSymbol(STR2, Symbol));
    Last = NULL;
    Symb = *Head;
    while (Symb != NULL) {
	Last = Symb;
	if (!strcmp(Symbol, Last->Symbol))
	    Result = false;
	if (strcmp(Symbol, Symb->Symbol) <= 0)
	    Symb = Symb->Left;
	else
	    Symb = Symb->Right;
    }
    Symb = (SymbolRecord *)Malloc(sizeof(SymbolRecord));
    strcpy(Symb->Symbol, Symbol);
    Symb->Address = Value;
    Symb->Left = NULL;
    Symb->Right = NULL;
    if (Last == NULL)
	*Head = Symb;
    else {
	if (strcmp(Symbol, Last->Symbol) <= 0)
	    Last->Left = Symb;
	else
	    Last->Right = Symb;
    }
    LastSymbol = Symb;
    return (Result);
}


Static boolean ResolveSymbol(Symbol_, Value, Head)
Char *Symbol_;
long *Value;
SymbolRecord *Head;
{
    boolean Result;
    SymbolString Symbol, STR2;

    strcpy(Symbol, Symbol_);
    Result = false;
    *Value = 0;
    strcpy(Symbol, StripSymbol(STR2, Symbol));
    while (Head != NULL) {
	if (!strcmp(Symbol, Head->Symbol)) {
	    *Value = Head->Address;
	    Result = true;
	    Head = NULL;
	} else {
	    if (strcmp(Symbol, Head->Symbol) < 0)
		Head = Head->Left;
	    else
		Head = Head->Right;
	}
    }
    return (Result);
}


Static boolean ResolveNumber(Symbol, Base, Value)
Char *Symbol;
uchar Base;
long *Value;
{
    boolean Result;
    uchar Digit, Count;
    Char STR2[256];

    Result = true;
    if (*Symbol == '\0')
	Result = false;
    *Value = 0;
    Count = 1;
    while (Count <= strlen(Symbol)) {
	sprintf(STR2, "%c", Symbol[Count - 1]);
	Digit = strpos2(DigitText, STR2, 1);
	if (Digit < 1 || Digit > Base) {
	    Result = false;
	    Count = strlen(Symbol);
	} else
	    *Value = Base * *Value + Digit - 1;
	Count++;
    }
    return (Result);
}


Static boolean ResolveValue(Symbol, Value, Head)
Char *Symbol;
long *Value;
SymbolRecord *Head;
{
    boolean Result;
    Char STR2[256];
    Char STR3[256];

    Result = false;
    *Value = 0;
    sprintf(STR2, "%c", AddressChar);
    if (!strcmp(Symbol, STR2)) {
	*Value = Current;
	return (true);
    }
    switch (Symbol[0]) {

    case HexChar:
	Result = ResolveNumber(strsub(STR3, Symbol, 2,
				      (int)(strlen(Symbol) - 1)), 16, Value);
	break;

    case BinChar:
	Result = ResolveNumber(strsub(STR3, Symbol, 2,
				      (int)(strlen(Symbol) - 1)), 2, Value);
	break;

    case OctChar:
	Result = ResolveNumber(strsub(STR3, Symbol, 2,
				      (int)(strlen(Symbol) - 1)), 8, Value);
	break;

    case AscChar:
	if (strlen(Symbol) >= 2)
	    *Value = Symbol[1];
	if (strlen(Symbol) == 2)
	    Result = true;
	break;

    default:
	if (isdigit(Symbol[0]))
	    Result = ResolveNumber(Symbol, 10, Value);
	else
	    Result = ResolveSymbol(Symbol, Value, Head);
	break;
    }
    return (Result);
}


Static Char *SplitParam(Result, Param)
Char *Result;
Char *Param;
{
    ParamString FoundWord;
    Char Character;
    uchar Start, Count;
    boolean Reading;
    Char STR2[256];

    *FoundWord = '\0';
    Reading = true;
    Start = 0;
    Count = 1;
    while (Count <= strlen(Param)) {
	Character = Param[Count - 1];
	if (Character > SpaceChar) {
	    Start = Count;
	    Count = strlen(Param);
	}
	Count++;
    }
    if (Start > 0) {
	Count = Start;
	Character = Param[Count - 1];
	if (P_inset(Character, Operators)) {
	    sprintf(FoundWord, "%c", Character);
	    strcpy(Param,
		   strsub(STR2, Param, Count + 1, strlen(Param) - Count));
/* p2c: chipper.pas, line 541:
 * Note: Possible string truncation in assignment [145] */
	    Reading = false;
	} else {
	    while (Count <= strlen(Param)) {
		Character = Param[Count - 1];
		if (Character <= SpaceChar || P_inset(Character, Operators)) {
		    strcpy(Param, strsub(STR2, Param, Count,
					 (int)(strlen(Param) - Count + 1)));
/* p2c: chipper.pas, line 550:
 * Note: Possible string truncation in assignment [145] */
		    Count = strlen(Param);
		    Reading = false;
		} else {
		    sprintf(FoundWord + strlen(FoundWord), "%c", Character);
/* p2c: chipper.pas, line 555:
 * Note: Possible string truncation in assignment [145] */
		}
		Count++;
	    }
	}
    }
    if (Reading)
	*Param = '\0';
    return (strcpy(Result, FoundWord));
/* p2c: chipper.pas, line 561:
 * Note: Possible string truncation in assignment [145] */
}


Static boolean ResolveOperator(Symbol, Token_, Legal)
Char *Symbol;
Char *Token_;
long *Legal;
{
    *Token_ = SpaceChar;
    if (strlen(Symbol) >= 1)
	*Token_ = Symbol[0];
    return (P_inset(*Token_, Legal));
}


Static boolean ResolveSingle(Token_, Value)
Char Token_;
long *Value;
{
    boolean Result;

    Result = true;
    switch (Token_) {

    case PlusChar:
	/* blank case */
	break;

    /* Nothing to be done */
    case MinusChar:
	*Value = -*Value;
	break;

    case NotChar:
	*Value = ~*Value;
	break;

    default:
	Result = false;
	break;
    }
    return (Result);
}


Static boolean ResolveDouble(Token_, Value, Operand)
Char Token_;
long *Value, Operand;
{
    boolean Result;
    long Count, This;

    Result = true;
    switch (Token_) {

    case PowerChar:
	This = 1;
	for (Count = 1; Count <= Operand; Count++)
	    This *= *Value;
	*Value = This;
	break;

    case ShlChar:
	*Value <<= Operand;
	break;

    case ShrChar:
	*Value = ((unsigned long)(*Value)) >> Operand;
	break;

    case MulChar:
	*Value *= Operand;
	break;

    case FracChar:
    case DivChar:
	*Value /= Operand;
	break;

    case PlusChar:
	*Value += Operand;
	break;

    case MinusChar:
	*Value -= Operand;
	break;

    case AndChar:
	*Value &= Operand;
	break;

    case OrChar:
	*Value |= Operand;
	break;

    case XorChar:
	*Value ^= Operand;
	break;

    case ModChar:
	*Value %= Operand;
/* p2c: chipper.pas, line 625:
 * Note: Using % for possibly-negative arguments [317] */
	break;

    default:
	Result = false;
	break;
    }
    return (Result);
}


Static boolean ResolveDivMod PP((Char *Symbol, Char *Param, long *Value,
				 SymbolRecord *Head));


Static boolean ResolveParent(Symbol, Param, Value, Head)
Char *Symbol;
Char *Param;
long *Value;
SymbolRecord *Head;
{
    Char Token_;
    boolean Status;
    long SET1[StartChar / 32 + 2];
    long SET2[StopChar / 32 + 2];

    if (!ResolveOperator(Symbol, &Token_,
			 P_addset(P_expset(SET1, 0L), StartChar))) {
	Status = ResolveValue(Symbol, Value, Head);
	SplitParam(Symbol, Param);
	return (Status);
    }
    SplitParam(Symbol, Param);
    Status = ResolveDivMod(Symbol, Param, Value, Head);
    if (!ResolveOperator(Symbol, &Token_,
			 P_addset(P_expset(SET2, 0L), StopChar)))
	Status = false;
    SplitParam(Symbol, Param);
    return (Status);
}


Static boolean ResolveUnary(Symbol, Param, Value, Head)
Char *Symbol;
Char *Param;
long *Value;
SymbolRecord *Head;
{
    Char Token_;
    boolean Status;
    long SET1[NotChar / 32 + 2];

    P_addset(P_expset(SET1, 0L), PlusChar);
    P_addset(SET1, MinusChar);
    if (!ResolveOperator(Symbol, &Token_, P_addset(SET1, NotChar))) {
	Status = ResolveParent(Symbol, Param, Value, Head);
	return (Status);
    }
    SplitParam(Symbol, Param);
    Status = ResolveParent(Symbol, Param, Value, Head);
    Status = (Status && ResolveSingle(Token_, Value));
    return (Status);
}


Static boolean ResolvePower(Symbol, Param, Value, Head)
Char *Symbol;
Char *Param;
long *Value;
SymbolRecord *Head;
{
    long Operand;
    Char Token_;
    boolean Status;
    long SET1[ShrChar / 32 + 2];

    Operand = 0;
    Status = ResolveUnary(Symbol, Param, Value, Head);
    while (ResolveOperator(Symbol, &Token_,
	       (P_addset(P_expset(SET1, 0L), PowerChar),
		P_addset(SET1, ShlChar), P_addset(SET1, ShrChar)))) {
	SplitParam(Symbol, Param);
	Status = (Status && ResolveUnary(Symbol, Param, &Operand, Head));
	Status = (Status && ResolveDouble(Token_, Value, Operand));
    }
    return (Status);
}


Static boolean ResolveMulDiv(Symbol, Param, Value, Head)
Char *Symbol;
Char *Param;
long *Value;
SymbolRecord *Head;
{
    long Operand;
    Char Token_;
    boolean Status;
    long SET1[FracChar / 32 + 2];

    Operand = 0;
    Status = ResolvePower(Symbol, Param, Value, Head);
    while (ResolveOperator(Symbol, &Token_,
			   (P_addset(P_expset(SET1, 0L), MulChar),
			    P_addset(SET1, FracChar)))) {
	SplitParam(Symbol, Param);
	Status = (Status && ResolvePower(Symbol, Param, &Operand, Head));
	Status = (Status && ResolveDouble(Token_, Value, Operand));
    }
    return (Status);
}


Static boolean ResolvePlusMinus(Symbol, Param, Value, Head)
Char *Symbol;
Char *Param;
long *Value;
SymbolRecord *Head;
{
    long Operand;
    Char Token_;
    boolean Status;
    long SET1[MinusChar / 32 + 2];

    Operand = 0;
    Status = ResolveMulDiv(Symbol, Param, Value, Head);
    while (ResolveOperator(Symbol, &Token_,
			   (P_addset(P_expset(SET1, 0L), PlusChar),
			    P_addset(SET1, MinusChar)))) {
	SplitParam(Symbol, Param);
	Status = (Status && ResolveMulDiv(Symbol, Param, &Operand, Head));
	Status = (Status && ResolveDouble(Token_, Value, Operand));
    }
    return (Status);
}


Static boolean ResolveBitWise(Symbol, Param, Value, Head)
Char *Symbol;
Char *Param;
long *Value;
SymbolRecord *Head;
{
    long Operand;
    Char Token_;
    boolean Status;
    long SET1[OrChar / 32 + 2];

    Operand = 0;
    Status = ResolvePlusMinus(Symbol, Param, Value, Head);
    while (ResolveOperator(Symbol, &Token_,
	       (P_addset(P_expset(SET1, 0L), AndChar), P_addset(SET1, OrChar),
		P_addset(SET1, XorChar)))) {
	SplitParam(Symbol, Param);
	Status = (Status && ResolvePlusMinus(Symbol, Param, &Operand, Head));
	Status = (Status && ResolveDouble(Token_, Value, Operand));
    }
    return (Status);
}


Static boolean ResolveDivMod(Symbol, Param, Value, Head)
Char *Symbol;
Char *Param;
long *Value;
SymbolRecord *Head;
{
    long Operand;
    Char Token_;
    boolean Status;
    long SET1[DivChar / 32 + 2];

    Operand = 0;
    Status = ResolveBitWise(Symbol, Param, Value, Head);
    while (ResolveOperator(Symbol, &Token_,
			   (P_addset(P_expset(SET1, 0L), DivChar),
			    P_addset(SET1, ModChar)))) {
	SplitParam(Symbol, Param);
	Status = (Status && ResolveBitWise(Symbol, Param, &Operand, Head));
	Status = (Status && ResolveDouble(Token_, Value, Operand));
    }
    return (Status);
}


Static boolean ResolveExpression(Param_, Value, Head)
Char *Param_;
long *Value;
SymbolRecord *Head;
{
    boolean Result;
    ParamString Param;
    SymbolString Symbol;

    strcpy(Param, Param_);
    Result = false;
    *Value = 0;
    SplitParam(Symbol, Param);
    if (*Symbol != '\0')
	Result = ResolveDivMod(Symbol, Param, Value, Head);
    if (*Symbol != '\0')
	return (false);
    return (Result);
}


Static Void StoreDirectives(Min, Max, DirHead)
uchar Min, Max;
SymbolRecord **DirHead;
{
    uchar This;

    This = Min + (Max - Min) / 2;
    if (!DefineSymbol(TokenText[(long)((Token)This)], (long)This, DirHead))
	RunWarning(CopySymbolWarning);
    if (This != Min)
	StoreDirectives(Min, This - 1, DirHead);
    if (This != Max)
	StoreDirectives(This + 1, Max, DirHead);
}


Static Void StoreRegisters(Min, Max, RegHead)
uchar Min, Max;
SymbolRecord **RegHead;
{
    uchar This;

    This = Min + (Max - Min) / 2;
    if (!DefineSymbol(RegisterText[(long)((Register_)This)], (long)This,
		      RegHead))
	RunWarning(CopySymbolWarning);
    if (This != Min)
	StoreRegisters(Min, This - 1, RegHead);
    if (This != Max)
	StoreRegisters(This + 1, Max, RegHead);
}


Static Char *SplitLine(Result, Line, AbortFlag)
Char *Result;
Char *Line;
boolean AbortFlag;
{
    ParamString FoundWord;
    Char Character;
    uchar Level, Start, Count;
    boolean TextFlag, Reading;
    Char STR2[256];

    *FoundWord = '\0';
    Level = 0;
    TextFlag = false;
    Reading = true;
    Start = 0;
    Count = 1;
    while (Count <= strlen(Line)) {
	Character = Line[Count - 1];
	if (Character == RemarkChar)
	    Count = strlen(Line);
	else {
	    if (Character > SpaceChar && Character != SeparatorChar) {
		Start = Count;
		Count = strlen(Line);
	    }
	}
	Count++;
    }
    if (Start > 0) {
	Count = Start;
	Count--;
	while (Count < strlen(Line)) {
	    Count++;
	    Character = Line[Count - 1];
	    if (Character == TextChar) {
		if (Line[Count - 2] == TextChar) {
		    sprintf(FoundWord + strlen(FoundWord), "%c", TextChar);
/* p2c: chipper.pas, line 864:
 * Note: Possible string truncation in assignment [145] */
		}
		TextFlag = !TextFlag;
		continue;
	    }
	    if (TextFlag) {
		sprintf(FoundWord + strlen(FoundWord), "%c", Character);
/* p2c: chipper.pas, line 870:
 * Note: Possible string truncation in assignment [145] */
		continue;
	    }
	    switch (Character) {

	    case SeparatorChar:
		if (Level == 0) {
		    strcpy(Line, strsub(STR2, Line, Count,
					(int)(strlen(Line) - Count + 1)));
		    Reading = false;
		    Count = strlen(Line);
		} else {
		    sprintf(FoundWord + strlen(FoundWord), "%c", Character);
/* p2c: chipper.pas, line 892:
 * Note: Possible string truncation in assignment [145] */
		}
		break;

	    case RemarkChar:
		*Line = '\0';
		Reading = false;
		Count = strlen(Line);
		break;

	    case StartChar:
		Level++;
		sprintf(FoundWord + strlen(FoundWord), "%c", Character);
/* p2c: chipper.pas, line 902:
 * Note: Possible string truncation in assignment [145] */
		break;

	    case StopChar:
		Level--;
		sprintf(FoundWord + strlen(FoundWord), "%c", Character);
/* p2c: chipper.pas, line 907:
 * Note: Possible string truncation in assignment [145] */
		break;

	    default:
		if ((uchar)Character <= SpaceChar) {
		    if (AbortFlag && Level == 0) {
			strcpy(Line, strsub(STR2, Line, Count,
					    (int)(strlen(Line) - Count + 1)));
			Reading = false;
			Count = strlen(Line);
		    } else {
			if (FoundWord[strlen(FoundWord) - 1] != SpaceChar) {
			    sprintf(FoundWord + strlen(FoundWord), "%c",
				    SpaceChar);
/* p2c: chipper.pas, line 882:
 * Note: Possible string truncation in assignment [145] */
			}
		    }
		} else {
		    sprintf(FoundWord + strlen(FoundWord), "%c",
			    toupper(Character));
/* p2c: chipper.pas, line 910:
 * Note: Possible string truncation in assignment [145] */
		}
		break;
	    }
	}
    }
    if (Reading)
	*Line = '\0';
    while (FoundWord[strlen(FoundWord) - 1] == SpaceChar)
	FoundWord[strlen(FoundWord) - 1] = '\0';
    return (strcpy(Result, FoundWord));
}


Static Void ReleaseSymbols(SymbHead)
SymbolRecord **SymbHead;
{
    if ((*SymbHead)->Left != NULL)
	ReleaseSymbols(&(*SymbHead)->Left);
    if ((*SymbHead)->Right != NULL)
	ReleaseSymbols(&(*SymbHead)->Right);
    Free(*SymbHead);
    *SymbHead = NULL;
}


Static Void ReleaseParams(ParamHead)
ParamRecord **ParamHead;
{
    ParamRecord *Param;

    while (*ParamHead != NULL) {
	Param = (*ParamHead)->Next;
	Free(*ParamHead);
	*ParamHead = Param;
    }
}


Static Void ReleaseInsts(InstHead)
InstRecord **InstHead;
{
    InstRecord *Inst;

    while (*InstHead != NULL) {
	ReleaseParams(&(*InstHead)->Params);
	Inst = (*InstHead)->Prev;
	Free(*InstHead);
	*InstHead = Inst;
    }
}


Static Void AlignWordBounds()
{
    Current = (Current + 1) / 2 * 2;
    if (Current > Finish)
	Finish = Current;
    if (Current < StartAddress || Current > StopAddress)
	RunError(BoundsError);
}


Static boolean ParamCheck(Count, Min, Max)
uchar Count, Min, Max;
{
    boolean Result;

    Result = false;
    if (Count < Min || Count > Max)
	RunWarning(ParamCountWarning);
    if (Count >= Min)
	return (true);
    return (Result);
}


Static boolean RangeCheck(Value, Min, Max, Message)
long Value, Min, Max;
Char *Message;
{
    boolean Result;

    Result = false;
    if (Value < Min || Value > Max) {
	RunWarning(Message);
	return (true);
    }
    return (Result);
}


Static Void DecodeFile PP((Char *Name, InstRecord **Head, SymbolRecord *Dir,
			   SymbolRecord **Symb));


Static uchar DecodeParameters(Line_, Head)
Char *Line_;
ParamRecord **Head;
{
    LineString Line;
    ParamString FoundWord;
    ParamRecord *Param, *Last;
    uchar Count;

    strcpy(Line, Line_);
    *Head = NULL;
    Count = 0;
    Last = NULL;
    do {
	SplitLine(FoundWord, Line, false);
	if (*FoundWord != '\0') {
	    Param = (ParamRecord *)Malloc(sizeof(ParamRecord));
	    strcpy(Param->Param, FoundWord);
	    Param->Next = NULL;
	    if (Last == NULL)
		*Head = Param;
	    else
		Last->Next = Param;
	    Last = Param;
	    Count++;
	}
    } while (*Line != '\0');
    return (Count);
}


Static Void DecodeDirective(Direct, Line, Valid, Head, Dir, Symb)
Token Direct;
Char *Line;
boolean Valid;
InstRecord **Head;
SymbolRecord *Dir, **Symb;
{
    InstRecord *Inst;
    ParamRecord *ParamHead;
    long Value;
    uchar Count;

    Count = DecodeParameters(Line, &ParamHead);
    switch (Direct) {

    case EquToken:
    case EqualToken:
	if (!Valid)
	    RunWarning(NoSymbolWarning);
	else {
	    if (ParamCheck(Count, 1, 1)) {
		if (!ResolveExpression(ParamHead->Param, &Value, *Symb))
		    RunWarning(UndefinedWarning);
		if (LastSymbol != NULL)
		    LastSymbol->Address = Value;
	    }
	}
	break;

    case DsToken:
	if (ParamCheck(Count, 1, 1)) {
	    if (!ResolveExpression(ParamHead->Param, &Value, *Symb))
		RunWarning(UndefinedWarning);
	    Current += Value;
	}
	break;

    case OrgToken:
	if (ParamCheck(Count, 1, 1)) {
	    if (ResolveExpression(ParamHead->Param, &Value, *Symb))
		Current = Value;
	    else
		RunWarning(UndefinedWarning);
	}
	break;

    case IncludeToken:
	if (ParamCheck(Count, 1, 1))
	    DecodeFile(ParamHead->Param, Head, Dir, Symb);
	ReleaseParams(&ParamHead);
	break;

    case EndToken:
	ParamCheck(Count, 0, 0);
	/* The END directive is ignored */
	ReleaseParams(&ParamHead);
	break;

    default:
	Inst = (InstRecord *)Malloc(sizeof(InstRecord));
	Inst->Line = LineNum;
	strcpy(Inst->Name, LineName);
	Inst->Address = Current;
	Inst->Inst = Direct;
	Inst->Count = Count;
	Inst->Params = ParamHead;
	Inst->Next = NULL;
	Inst->Prev = *Head;
	if (*Head != NULL)
	    (*Head)->Next = Inst;
	*Head = Inst;
	switch (Direct) {

	case DbToken:
	    if (ParamCheck(Count, 1, MaxParams))
		Current += Count;
	    break;

	case DwToken:
	    if (ParamCheck(Count, 1, MaxParams))
		Current += Count * 2;
	    break;

	case DaToken:
	    if (ParamCheck(Count, 1, 1))
		Current += strlen(ParamHead->Param);
	    break;

	default:
	    Current += 2;
	    break;
	}
	break;
    }
}


Static Void DecodeLine(Line_, Head, Dir, Symb)
Char *Line_;
InstRecord **Head;
SymbolRecord *Dir, **Symb;
{
    LineString Line;
    ParamString Direct;
    long Value;
    boolean Valid;

    strcpy(Line, Line_);
    Valid = false;
    do {
	SplitLine(Direct, Line, true);
	if (*Direct != '\0') {
	    if (!ResolveSymbol(Direct, &Value, Dir)) {
		if (Valid)
		    RunWarning(DualSymbolWarning);
		if (!DefineSymbol(Direct, Current, Symb))
		    RunWarning(CopySymbolWarning);
		Valid = true;
	    } else {
		DecodeDirective((Token)Value, Line, Valid, Head, Dir, Symb);
		AlignWordBounds();
		*Line = '\0';
	    }
	}
    } while (*Line != '\0');
}


Static Void DecodeFile(Name, Head, Dir, Symb)
Char *Name;
InstRecord **Head;
SymbolRecord *Dir, **Symb;
{
    FILE *InFile;
    LineString Line;
    Char StoreName[CHIP_PATH_LEN];
    unsigned short StoreNum;
    Char *TEMP;
    Char InFile_NAME[_FNSIZE];

    InFile = NULL;
    strcpy(StoreName, LineName);
    StoreNum = LineNum;
    ExpandFileName(LineName, Name, NameDefault, InExtDefault);
    LineNum = 1;
    fprintf(StdOut, "Reading: %s\n\n", LineName);
    fprintf(ListFile, "Reading: %s\n\n", LineName);
    strcpy(InFile_NAME, LineName);
    if (InFile != NULL)
	InFile = freopen(InFile_NAME, "r", InFile);
    else
	InFile = fopen(InFile_NAME, "r");
    _SETIO(InFile != NULL, FileNotFound);
    if (P_ioresult != 0)
	RunError(FileAccessError);
    while (fgets(Line, LineLength + 1, InFile) != NULL) {
	TEMP = strchr(Line, '\n');
	if (TEMP != NULL)
	    *TEMP = 0;
	strcpy(LineText, Line);
	DecodeLine(Line, Head, Dir, Symb);
	LineNum++;
    }
    if (InFile != NULL)
	fclose(InFile);
    InFile = NULL;
    strcpy(LineName, StoreName);
    LineNum = StoreNum;
    if (strcmp(LineName, NameDefault)) {
	fprintf(StdOut, "Reading: %s\n\n", LineName);
	fprintf(ListFile, "Reading: %s\n\n", LineName);
    } else {
	fprintf(StdOut, "Done reading\n\n");
	fprintf(ListFile, "Done reading\n\n");
    }
    if (InFile != NULL)
	fclose(InFile);
}


Static Void EncodeNoneToken(Instruct, OpCode, Reg, Symb)
InstRecord *Instruct;
unsigned short OpCode;
SymbolRecord *Reg, *Symb;
{
    ParamCheck(Instruct->Count, 0, 0);
    /* Generate instruction anyway */
    Memory[Instruct->Address - StartAddress] = OpCode / 256;
    Memory[Instruct->Address - StartAddress + 1] = OpCode & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 0);
}


Static Void EncodeAddrToken(Instruct, OpCode, Reg, Symb)
InstRecord *Instruct;
unsigned short OpCode;
SymbolRecord *Reg, *Symb;
{
    long Addr;

    Addr = 0;
    if (ParamCheck(Instruct->Count, 1, 1)) {
	if (!ResolveExpression(Instruct->Params->Param, &Addr, Symb))
	    RunWarning(UndefinedWarning);
    }
    if (RangeCheck(Addr, 0L, (long)AddrMask, RangeWarning))
	Addr &= AddrMask;
    Addr |= OpCode;
    Memory[Instruct->Address - StartAddress] = Addr / 256;
    Memory[Instruct->Address - StartAddress + 1] = Addr & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 1);
}


Static Void EncodeRegToken(Instruct, OpCode, Reg, Symb)
InstRecord *Instruct;
unsigned short OpCode;
SymbolRecord *Reg, *Symb;
{
    long RegX;

    RegX = (long)V0Reg;
    if (ParamCheck(Instruct->Count, 1, 1)) {
	if (ResolveSymbol(Instruct->Params->Param, &RegX, Reg))
	    RegX -= (int)V0Reg;
	else
	    RunWarning(NoRegisterWarning);
    }
    if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	RegX = 0;
    RegX = OpCode | (RegX * 0x100);
    Memory[Instruct->Address - StartAddress] = RegX / 256;
    Memory[Instruct->Address - StartAddress + 1] = RegX & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 1);
}


Static Void EncodeRegValToken(Instruct, OpCode, Reg, Symb)
InstRecord *Instruct;
unsigned short OpCode;
SymbolRecord *Reg, *Symb;
{
    long RegX, Value;

    RegX = (long)V0Reg;
    Value = 0;
    if (Instruct->Count >= 1) {
	if (ResolveSymbol(Instruct->Params->Param, &RegX, Reg))
	    RegX -= (int)V0Reg;
	else
	    RunWarning(NoRegisterWarning);
    }
    if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	RegX = 0;
    if (ParamCheck(Instruct->Count, 2, 2)) {
	if (!ResolveExpression(Instruct->Params->Next->Param, &Value, Symb))
	    RunWarning(UndefinedWarning);
    }
    if (RangeCheck(Value, 0L, (long)ByteMask, RangeWarning))
	Value &= ByteMask;
    Value |= OpCode | (RegX * 0x100);
    Memory[Instruct->Address - StartAddress] = Value / 256;
    Memory[Instruct->Address - StartAddress + 1] = Value & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 2);
}


Static Void EncodeRegRegToken(Instruct, OpCode, Min, Reg, Symb)
InstRecord *Instruct;
unsigned short OpCode;
uchar Min;
SymbolRecord *Reg, *Symb;
{
    long RegX, RegY;

    RegX = (long)V0Reg;
    RegY = (long)V0Reg;
    if (ParamCheck(Instruct->Count, Min, 2)) {
	if (ResolveSymbol(Instruct->Params->Param, &RegX, Reg))
	    RegX -= (int)V0Reg;
	else
	    RunWarning(NoRegisterWarning);
    }
    if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	RegX = 0;
    if (Instruct->Count >= 2) {
	if (ResolveSymbol(Instruct->Params->Next->Param, &RegY, Reg))
	    RegY -= (int)V0Reg;
	else
	    RunWarning(NoRegisterWarning);
    }
    if (RangeCheck(RegY, 0L, (long)NibbleMask, BadRegisterWarning))
	RegY = 0;
    RegX = OpCode | (RegX * 0x100) | (RegY * 0x10);
    Memory[Instruct->Address - StartAddress] = RegX / 256;
    Memory[Instruct->Address - StartAddress + 1] = RegX & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 2);
}


Static Void EncodeRegRegOrValToken(Instruct, OpCode1, OpCode2, Reg, Symb)
InstRecord *Instruct;
unsigned short OpCode1, OpCode2;
SymbolRecord *Reg, *Symb;
{
    long RegX, RegY, Value;

    RegX = (long)V0Reg;
    RegY = (long)V0Reg;
    Value = 0;
    if (Instruct->Count >= 1) {
	if (ResolveSymbol(Instruct->Params->Param, &RegX, Reg))
	    RegX -= (int)V0Reg;
	else
	    RunWarning(NoRegisterWarning);
    }
    if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	RegX = 0;
    if (!ParamCheck(Instruct->Count, 2, 2))
	return;
    if (ResolveSymbol(Instruct->Params->Next->Param, &RegY, Reg)) {
	RegY -= (int)V0Reg;
	if (RangeCheck(RegY, 0L, (long)NibbleMask, BadRegisterWarning))
	    RegY = 0;
	Value = OpCode1 | (RegX * 0x100) | (RegY * 0x10);
	Memory[Instruct->Address - StartAddress] = Value / 256;
	Memory[Instruct->Address - StartAddress + 1] = Value & 255;
	ListInstruction((int)Instruct->Address, 2, Instruct, 2);
	return;
    }
    if (!ResolveExpression(Instruct->Params->Next->Param, &Value, Symb))
	RunWarning(UndefinedWarning);
    if (RangeCheck(Value, 0L, (long)ByteMask, RangeWarning))
	Value &= ByteMask;
    Value |= OpCode2 | (RegX * 0x100);
    Memory[Instruct->Address - StartAddress] = Value / 256;
    Memory[Instruct->Address - StartAddress + 1] = Value & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 2);
}


Static Void EncodeRegRegValToken(Instruct, OpCode, Reg, Symb)
InstRecord *Instruct;
unsigned short OpCode;
SymbolRecord *Reg, *Symb;
{
    long RegX, RegY, Value;

    RegX = (long)V0Reg;
    RegY = (long)V0Reg;
    Value = 0;
    if (Instruct->Count >= 1) {
	if (ResolveSymbol(Instruct->Params->Param, &RegX, Reg))
	    RegX -= (int)V0Reg;
	else
	    RunWarning(NoRegisterWarning);
    }
    if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	RegX = 0;
    if (Instruct->Count >= 2) {
	if (ResolveSymbol(Instruct->Params->Next->Param, &RegY, Reg))
	    RegY -= (int)V0Reg;
	else
	    RunWarning(NoRegisterWarning);
    }
    if (RangeCheck(RegY, 0L, (long)NibbleMask, BadRegisterWarning))
	RegY = 0;
    if (ParamCheck(Instruct->Count, 3, 3)) {
	if (!ResolveExpression(Instruct->Params->Next->Next->Param, &Value,
			       Symb))
	    RunWarning(UndefinedWarning);
    }
    if (RangeCheck(Value, 0L, (long)NibbleMask, RangeWarning))
	Value &= NibbleMask;
    Value |= OpCode | (RegX * 0x100) | (RegY * 0x10);
    Memory[Instruct->Address - StartAddress] = Value / 256;
    Memory[Instruct->Address - StartAddress + 1] = Value & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 3);
}


Static Void EncodeAddToken(Instruct, Reg, Symb)
InstRecord *Instruct;
SymbolRecord *Reg, *Symb;
{
    long RegX, RegY, Value;

    RegX = (long)V0Reg;
    RegY = (long)V0Reg;
    Value = 0;
    if (Instruct->Count >= 1) {
	if (!ResolveSymbol(Instruct->Params->Param, &RegX, Reg))
	    RunWarning(NoRegisterWarning);
    }
    if (RegX == (int)IReg) {
	if (ParamCheck(Instruct->Count, 2, 2)) {
	    if (ResolveSymbol(Instruct->Params->Next->Param, &RegX, Reg))
		RegX -= (int)V0Reg;
	    else
		RunWarning(NoRegisterWarning);
	}
	if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	    RegX = 0;
	Value = 0xf01eL | (RegX * 0x100);
	Memory[Instruct->Address - StartAddress] = Value / 256;
	Memory[Instruct->Address - StartAddress + 1] = Value & 255;
	ListInstruction((int)Instruct->Address, 2, Instruct, 2);
	return;
    }
    RegX -= (int)V0Reg;
    if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	RegX = 0;
    if (!ParamCheck(Instruct->Count, 2, 2))
	return;
    if (ResolveSymbol(Instruct->Params->Next->Param, &RegY, Reg)) {
	RegY -= (int)V0Reg;
	if (RangeCheck(RegY, 0L, (long)NibbleMask, BadRegisterWarning))
	    RegY = 0;
	Value = 0x8004L | (RegX * 0x100) | (RegY * 0x10);
	Memory[Instruct->Address - StartAddress] = Value / 256;
	Memory[Instruct->Address - StartAddress + 1] = Value & 255;
	ListInstruction((int)Instruct->Address, 2, Instruct, 2);
	return;
    }
    if (!ResolveExpression(Instruct->Params->Next->Param, &Value, Symb))
	RunWarning(UndefinedWarning);
    if (RangeCheck(Value, 0L, (long)ByteMask, RangeWarning))
	Value &= ByteMask;
    Value |= 0x7000 | (RegX * 0x100);
    Memory[Instruct->Address - StartAddress] = Value / 256;
    Memory[Instruct->Address - StartAddress + 1] = Value & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 2);
}


Static Void EncodeJpToken(Instruct, Reg, Symb)
InstRecord *Instruct;
SymbolRecord *Reg, *Symb;
{
    long RegX, Addr;

    RegX = (long)V0Reg;
    Addr = (long)V0Reg;
    if (!ParamCheck(Instruct->Count, 1, 2))
	return;
    if (ResolveSymbol(Instruct->Params->Param, &RegX, Reg)) {
	if (RegX != (int)V0Reg)
	    RunWarning(BadRegisterWarning);
	if (Instruct->Count == 1)
	    RunWarning(ParamCountWarning);
	else {
	    if (!ResolveExpression(Instruct->Params->Next->Param, &Addr, Symb))
		RunWarning(UndefinedWarning);
	}
	if (RangeCheck(Addr, 0L, (long)AddrMask, RangeWarning))
	    Addr &= AddrMask;
	Addr |= 0xb000L;
	Memory[Instruct->Address - StartAddress] = Addr / 256;
	Memory[Instruct->Address - StartAddress + 1] = Addr & 255;
	ListInstruction((int)Instruct->Address, 2, Instruct, 2);
	return;
    }
    if (!ResolveExpression(Instruct->Params->Param, &Addr, Symb))
	RunWarning(UndefinedWarning);
    if (Instruct->Count == 2)
	RunWarning(ParamCountWarning);
    if (RangeCheck(Addr, 0L, (long)AddrMask, RangeWarning))
	Addr &= AddrMask;
    Addr |= 0x1000;
    Memory[Instruct->Address - StartAddress] = Addr / 256;
    Memory[Instruct->Address - StartAddress + 1] = Addr & 255;
    ListInstruction((int)Instruct->Address, 2, Instruct, 1);
}


Static Void EncodeLdToken(Instruct, Reg, Symb)
InstRecord *Instruct;
SymbolRecord *Reg, *Symb;
{
    long RegX, RegY, Value;
    boolean RegFlag;

    RegX = (long)V0Reg;
    RegY = (long)V0Reg;
    Value = 0;
    RegFlag = true;
    if (Instruct->Count >= 1) {
	if (!ResolveSymbol(Instruct->Params->Param, &RegX, Reg))
	    RunWarning(NoRegisterWarning);
    }
    switch (RegX) {

    case (long)BReg:
    case (long)DtReg:
    case (long)FReg:
    case (long)StReg:
    case (long)IiReg:
	RegY = RegX;
	RegX = (long)V0Reg;
	if (ParamCheck(Instruct->Count, 2, 2)) {
	    if (ResolveSymbol(Instruct->Params->Next->Param, &RegX, Reg))
		RegX -= (int)V0Reg;
	    else
		RunWarning(NoRegisterWarning);
	}
	if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	    RegX = 0;
	switch (RegY) {

	case (long)BReg:
	    Value = 0xf033L | (RegX * 0x100);
	    break;

	case (long)DtReg:
	    Value = 0xf015L | (RegX * 0x100);
	    break;

	case (long)FReg:
	    Value = 0xf029L | (RegX * 0x100);
	    break;

	case (long)StReg:
	    Value = 0xf018L | (RegX * 0x100);
	    break;

	case (long)IiReg:
	    Value = 0xf055L | (RegX * 0x100);
	    break;

	default:
	    RunWarning(InternalWarning);
	    break;
	}
	Memory[Instruct->Address - StartAddress] = Value / 256;
	Memory[Instruct->Address - StartAddress + 1] = Value & 255;
	ListInstruction((int)Instruct->Address, 2, Instruct, 2);
	break;

    case (long)IReg:
	if (ParamCheck(Instruct->Count, 2, 2)) {
	    if (!ResolveExpression(Instruct->Params->Next->Param, &Value,
				   Symb))
		RunWarning(UndefinedWarning);
	}
	if (RangeCheck(Value, 0L, (long)AddrMask, RangeWarning))
	    Value &= AddrMask;
	Value |= 0xa000L;
	Memory[Instruct->Address - StartAddress] = Value / 256;
	Memory[Instruct->Address - StartAddress + 1] = Value & 255;
	ListInstruction((int)Instruct->Address, 2, Instruct, 2);
	break;

    default:
	RegX -= (int)V0Reg;
	if (RangeCheck(RegX, 0L, (long)NibbleMask, BadRegisterWarning))
	    RegX = 0;
	if (ParamCheck(Instruct->Count, 2, 2))
	    RegFlag = ResolveSymbol(Instruct->Params->Next->Param, &RegY, Reg);
	if (RegFlag) {
	    switch (RegY) {

	    case (long)DtReg:
		Value = 0xf007L | (RegX * 0x100);
		Memory[Instruct->Address - StartAddress] = Value / 256;
		Memory[Instruct->Address - StartAddress + 1] = Value & 255;
		ListInstruction((int)Instruct->Address, 2, Instruct, 2);
		break;

	    case (long)KReg:
		Value = 0xf00aL | (RegX * 0x100);
		Memory[Instruct->Address - StartAddress] = Value / 256;
		Memory[Instruct->Address - StartAddress + 1] = Value & 255;
		ListInstruction((int)Instruct->Address, 2, Instruct, 2);
		break;

	    case (long)IiReg:
		Value = 0xf065L | (RegX * 0x100);
		Memory[Instruct->Address - StartAddress] = Value / 256;
		Memory[Instruct->Address - StartAddress + 1] = Value & 255;
		ListInstruction((int)Instruct->Address, 2, Instruct, 2);
		break;

	    default:
		RegY -= (int)V0Reg;
		if (RangeCheck(RegY, 0L, (long)NibbleMask, BadRegisterWarning))
		    RegY = 0;
		Value = 0x8000L | (RegX * 0x100) | (RegY * 0x10);
		Memory[Instruct->Address - StartAddress] = Value / 256;
		Memory[Instruct->Address - StartAddress + 1] = Value & 255;
		ListInstruction((int)Instruct->Address, 2, Instruct, 2);
		break;
	    }
	} else {
	    if (Instruct->Count >= 2) {
		if (!ResolveExpression(Instruct->Params->Next->Param, &Value,
				       Symb))
		    RunWarning(UndefinedWarning);
	    }
	    if (RangeCheck(Value, 0L, (long)ByteMask, RangeWarning))
		Value &= ByteMask;
	    Value |= 0x6000 | (RegX * 0x100);
	    Memory[Instruct->Address - StartAddress] = Value / 256;
	    Memory[Instruct->Address - StartAddress + 1] = Value & 255;
	    ListInstruction((int)Instruct->Address, 2, Instruct, 2);
	}
	break;
    }
}


Static Void EncodeDaToken(Instruct, Reg, Symb)
InstRecord *Instruct;
SymbolRecord *Reg, *Symb;
{
    ParamString Param;
    uchar This, FORLIM;

    *Param = '\0';
    if (Instruct->Count >= 1)
	strcpy(Param, Instruct->Params->Param);
    FORLIM = strlen(Param);
    for (This = 0; This < FORLIM; This++)
	Memory[Instruct->Address + This - StartAddress] = Param[This];
    ListInstruction((int)Instruct->Address, strlen(Param), Instruct, 1);
}


Static Void EncodeDbToken(Instruct, Reg, Symb)
InstRecord *Instruct;
SymbolRecord *Reg, *Symb;
{
    ParamRecord *Param;
    long Value;
    uchar This;

    This = 0;
    Value = 0;
    Param = Instruct->Params;
    while (Param != NULL) {
	if (!ResolveExpression(Param->Param, &Value, Symb))
	    RunWarning(UndefinedWarning);
	if (RangeCheck(Value, 0L, (long)ByteMask, RangeWarning))
	    Value &= ByteMask;
	Memory[Instruct->Address + This - StartAddress] = Value;
	This++;
	Param = Param->Next;
    }
    ListInstruction((int)Instruct->Address, Instruct->Count, Instruct,
		    Instruct->Count);
}


Static Void EncodeDwToken(Instruct, Reg, Symb)
InstRecord *Instruct;
SymbolRecord *Reg, *Symb;
{
    ParamRecord *Param;
    long Value;
    uchar This;

    This = 0;
    Value = 0;
    Param = Instruct->Params;
    while (Param != NULL) {
	if (!ResolveExpression(Param->Param, &Value, Symb))
	    RunWarning(UndefinedWarning);
	if (RangeCheck(Value, 0L, WordMask, RangeWarning))
	    Value &= WordMask;
	Memory[Instruct->Address + This - StartAddress] = Value / 256;
	Memory[Instruct->Address + This - StartAddress + 1] = Value & 255;
	This += 2;
	Param = Param->Next;
    }
    ListInstruction((int)Instruct->Address, Instruct->Count * 2, Instruct,
		    Instruct->Count);
}


Static Void EncodeInstruction(Inst, Reg, Symb)
InstRecord *Inst;
SymbolRecord *Reg, *Symb;
{
    switch (Inst->Inst) {

    case AddToken:
	EncodeAddToken(Inst, Reg, Symb);
	break;

    case AndToken:
	EncodeRegRegToken(Inst, (int)0x8002L, 2, Reg, Symb);
	break;

    case CallToken:
	EncodeAddrToken(Inst, 0x2000, Reg, Symb);
	break;

    case ClsToken:
	EncodeNoneToken(Inst, 0xe0, Reg, Symb);
	break;

    case DaToken:
	EncodeDaToken(Inst, Reg, Symb);
	break;

    case DbToken:
	EncodeDbToken(Inst, Reg, Symb);
	break;

    case DrwToken:
	EncodeRegRegValToken(Inst, (int)0xd000L, Reg, Symb);
	break;

    case DwToken:
	EncodeDwToken(Inst, Reg, Symb);
	break;

    case JpToken:
	EncodeJpToken(Inst, Reg, Symb);
	break;

    case LdToken:
	EncodeLdToken(Inst, Reg, Symb);
	break;

    case OrToken:
	EncodeRegRegToken(Inst, (int)0x8001L, 2, Reg, Symb);
	break;

    case RetToken:
	EncodeNoneToken(Inst, 0xee, Reg, Symb);
	break;

    case RndToken:
	EncodeRegValToken(Inst, (int)0xc000L, Reg, Symb);
	break;

    case SeToken:
	EncodeRegRegOrValToken(Inst, 0x5000, 0x3000, Reg, Symb);
	break;

    case ShlToken:
	EncodeRegRegToken(Inst, (int)0x800eL, 1, Reg, Symb);
	break;

    case ShrToken:
	EncodeRegRegToken(Inst, (int)0x8006L, 1, Reg, Symb);
	break;

    case SknpToken:
	EncodeRegToken(Inst, (int)0xe0a1L, Reg, Symb);
	break;

    case SkpToken:
	EncodeRegToken(Inst, (int)0xe09eL, Reg, Symb);
	break;

    case SneToken:
	EncodeRegRegOrValToken(Inst, (int)0x9000L, 0x4000, Reg, Symb);
	break;

    case SubToken:
	EncodeRegRegToken(Inst, (int)0x8005L, 2, Reg, Symb);
	break;

    case SubnToken:
	EncodeRegRegToken(Inst, (int)0x8007L, 2, Reg, Symb);
	break;

    case SysToken:
	EncodeAddrToken(Inst, 0x2000, Reg, Symb);
	break;

    case XorToken:
	EncodeRegRegToken(Inst, (int)0x8003L, 2, Reg, Symb);
	break;

    default:
	RunWarning(InternalWarning);
	break;
    }
}


Static Void EncodeMemory(Inst, Reg, Symb)
InstRecord *Inst;
SymbolRecord *Reg, *Symb;
{
    InstRecord *Prev;
    unsigned short Count;

    for (Count = StartAddress; Count <= StopAddress; Count++)
	Memory[Count - StartAddress] = 0;
    Prev = Inst;
    while (Prev != NULL) {
	Inst = Prev;
	Prev = Prev->Prev;
    }
    while (Inst != NULL) {
	InstPoint = Inst;
	Current = Inst->Address;
	EncodeInstruction(Inst, Reg, Symb);
	Inst = Inst->Next;
    }
    InstPoint = NULL;
    ListSymbols(Symb);
    ListWarnings();
}


Static Void WriteMemory(Start, Stop)
unsigned short Start, Stop;
{
    uchar HpHeading[0xd];
    long Size, Count;

    Size = (Stop - Start) * 2 + 5;
    HpHeading[0] = 'H';
    HpHeading[0x1] = 'P';
    HpHeading[0x2] = 'H';
    HpHeading[0x3] = 'P';
    HpHeading[0x4] = '4';
    HpHeading[0x5] = '8';
    HpHeading[0x6] = '-';
    HpHeading[0x7] = 'A';
    HpHeading[0x8] = 0x2c;
    HpHeading[0x9] = 0x2a;
    HpHeading[0xa] = 0 | ((Size & 15) * 0x10);
    HpHeading[0xb] = (Size / 16) & 255;
    HpHeading[0xc] = (Size / 4096) & 255;
    for (Count = 0; Count <= 0xc; Count++)
	fwrite(&HpHeading[Count], sizeof(uchar), 1, OutFile);
    for (Count = Start; Count < Stop; Count++)
	fwrite(&Memory[Count - StartAddress], sizeof(uchar), 1, OutFile);
}


main(argc, argv)
int argc;
Char *argv[];
{
Operators = P_addset(
		   P_addset(
			    P_addset(
				     P_addset(P_addset(
					     P_addset(
						      P_addset(P_addset(P_addset(P_addset(
											  P_addset(
P_addset(
P_addset(P_addset(P_addset(P_expset(
SET, 0L), StartChar), StopChar), PlusChar),
MinusChar), NotChar), PowerChar),
									    ShlChar), ShrChar), MulChar), FracChar),
					     AndChar),
					 OrChar), XorChar),
			    DivChar),
		   ModChar);
    PASCAL_MAIN(argc, argv);
    ListFile = NULL;
    StdOut = NULL;
    StdIn = NULL;
    OutFile = NULL;
    strcpy(StdIn_NAME, "/dev/tty");
    strcpy(StdOut_NAME, "/dev/tty");
    if (StdIn != NULL)
	StdIn = freopen(StdIn_NAME, "r", StdIn);
    else
	StdIn = fopen(StdIn_NAME, "r");
    if (StdIn == NULL)
	_EscIO(FileNotFound);
    if (StdOut != NULL)
	StdOut = freopen(StdOut_NAME, "w", StdOut);
    else
	StdOut = fopen(StdOut_NAME, "w");
    if (StdOut == NULL)
	_EscIO(FileNotFound);
    fprintf(StdOut, "%s\n\n", CopyRight);
    if (OpenFiles()) {
	StoreDirectives(0, (int)((Token)((long)LastToken - 1)), &Directives);
	StoreRegisters(0, (int)((Register_)((long)LastReg - 1)), &Registers_);
	DecodeFile(InFileName, &Instructions, Directives, &Symbols);
	EncodeMemory(Instructions, Registers_, Symbols);
	WriteMemory(StartAddress, (int)Finish);
	if (OutFile != NULL)
	    fclose(OutFile);
	OutFile = NULL;
	if (ListFile != NULL)
	    fclose(ListFile);
	ListFile = NULL;
	ReleaseSymbols(&Symbols);
	ReleaseSymbols(&Registers_);
	ReleaseSymbols(&Directives);
	ReleaseInsts(&Instructions);
    }
    if (StdIn != NULL)
	fclose(StdIn);
    StdIn = NULL;
    if (StdOut != NULL)
	fclose(StdOut);
    StdOut = NULL;
    if (OutFile != NULL)
	fclose(OutFile);
    if (StdIn != NULL)
	fclose(StdIn);
    if (StdOut != NULL)
	fclose(StdOut);
    if (ListFile != NULL)
	fclose(ListFile);
    exit(EXIT_SUCCESS);
}



/* End. */
